#include <iostream>
#include <cmath>
#include <cstring>
#include <cstdio>
#include <fstream>
#include "my_functions.h"

using namespace std;

const int Nz = 201;
const int NN = 25;
const double Ra = 100000;
const double Pr = 2;
const double a = 2;
const int Nx = 401;

int main(int argc, char * argv[])
{
  double Tem_v[NN][Nz];
  double ome_v[NN][Nz];
  double psi_v[NN][Nz];
  double dTem_v[NN][Nz];
  double dome_v[NN][Nz];
  double dpsi_v[NN][Nz];
  double dTemO_v[NN][Nz];
  double domeO_v[NN][Nz];
  double dpsiO_v[NN][Nz];
  double x[Nx];
  double z[Nz];
  double vx[Nx][Nz];
  double vz[Nx][Nz];
  double tem_mov[Nx][Nz];
  double psi_mov[Nx][Nz];
  double sina[NN][Nx];
  double cosa[NN][Nx];
  double npoa_cosa[NN][Nx];

  double time = 0.;
  double time_final = 0.25;
  double dt;
  double dx;
  double dz;

  double dia[NN][Nz];
  double sup[Nz];
  double sub[Nz];
  double tmp_var;
  double max_vel;

  int i,j,l,m;
  long istep = 0;
  long iprint = 200;
  char fname_stem[7];
  char fname_num[10];
  char fname[20];
  
  ofstream output_file;

  // Initialize vectors here
  strcpy(fname_stem, "mov.");
  
  dz = 1. / (Nz-1);
  dx = dz;
  for(j=0; j<Nz;j++)
    {
      z[j] = dz * (double) j;
    }
  for(j=0; j < Nx; j++)
    {
      x[j] = dx * (double) j;
    }

  for(j=0; j < NN; j++)
    {
      for(i=0; i < Nx; i++)
	{
	  sina[j][i]=sin((double) j * M_PI*x[i]/a);
	  cosa[j][i]=cos((double) j * M_PI*x[i]/a);
	  npoa_cosa[j][i] = ((double) j)* M_PI * cosa[j][i] / a;
	}
    }

  // Compute components for the tridiagonal solver
  for(i=0; i<NN; i++)
    {
      dia[i][0] = 1.;
      dia[i][Nz-1] = 1.;
      for(j=1; j < Nz-1; j++)
	{
	  dia[i][j] = pow((i*M_PI/a),2)+2/(dz*dz);	  
	}
    }
  for(j=0; j < Nz; j++)
    {
      sup[j] = -1. / (dz*dz);
      sub[j] = -1. / (dz*dz);
    }

  for(j=0; j < Nz; j++)
    {      
      for(i=0; i < NN; i++)
	{
	  // setup initial temperature perturbations
	  if(i==0)
	    {
	      Tem_v[i][j] = 1. - z[j];
	    }
	  else
	    {
	      if((j==0) || (j==Nz-1))
		{
		  Tem_v[i][j] = 0.;
		}
	      else
		{
		  Tem_v[i][j] = sin(M_PI*z[j]) / ((double) (i+1));
		}
	    }
	  ome_v[i][j] = 0.;
	  psi_v[i][j] = 0.;
	  // Initialize derivatives
	  dTem_v[i][j] = 0.;
	  dome_v[i][j] = 0.;
	  // Old derivatives
	  dTemO_v[i][j] = 0.;
	  domeO_v[i][j] = 0.;
	}
    }
  
  // loop here
  dt = (dz*dz) / (10.*Pr);

  // MAIN Loop
  while(time < time_final)
    {
      istep++;
      // Compute psi for current state
      for(i=1; i < NN; i++)
	{
	  tridiag(Nz, sub, dia[i], sup, ome_v[i], psi_v[i]);
	}

      // Diagonostic: Print out all of the first order modes
//      for(j=0; j < Nz; j++)
//	{
//	  cout << z[j] << ' ' << psi_v[1][j] << ' ' << Tem_v[1][j] << ' ' << ome_v[1][j] << endl;	    
//	}

      if(istep % iprint == 0)
	{
	  // compute velocities
	  // Start by reseting velocity values to zero
	  for(i=0; i < Nx; i++)
	    {
	      for(j=0; j < Nz ; j++)
		{
		  vx[i][j] = 0.;
		  vz[i][j] = 0.;
		  tem_mov[i][j] = 0.;
		  psi_mov[i][j] = 0.;
		}
	    }
	  for(j = 1 ; j < Nz-1; j++)
	    {
	      for(i=0; i < Nx; i++)
		{
		  for(l=1; l < NN; l++)
		    {
		      vx[i][j] += -((psi_v[l][j+1]-psi_v[l][j-1])/(2.*dz))
			* sina[l][i];
		      vz[i][j] += psi_v[l][j] * npoa_cosa[l][i];
		    }
		}
	    }

	  for(j = 0 ; j < Nz; j++)
	    {
	      for(i=0; i < Nx; i++)
		{
		  for(l=0; l < NN; l++)
		    {
		      tem_mov[i][j] += Tem_v[l][j]*cosa[l][i];
		      psi_mov[i][j] += psi_v[l][j]*sina[l][i];
		    }
		}
	    }
	  // Calculate dt
	  max_vel = 0.;
	  for(i=1; i < Nx-1; i++)
	    {
	      for(j=1; j < Nz-1; j++)
		{
		  
		  max_vel = max(max_vel, sqrt(vx[i][j]*vx[i][j]+vz[i][j]*vz[i][j]));
		}
	    }
	  dt = min(dz / max_vel / 10., (dz*dz) / (10.*Pr));
	  cout << time << ' ' << istep << ' ' << max_vel << ' ' << dt << ' ' << (dz*dz) / (10.*Pr) <<  endl;	  

	  // write to file
	  // What is the name of the file?
	  sprintf(fname_num, "%06i",istep / iprint);
	  strcpy(fname, fname_stem);	  
	  strcat(fname,fname_num);
	  // open the file
	  output_file.open(fname, ios::out);  
	  output_file << Nx << ' ' << Nz << ' ' << a << ' ' << Ra << ' ' << Pr << endl;
	  // write to the file
	  for(j=0; j < Nz; j++)
	    {
	      for(i=0; i < Nx; i++)
		{
		  output_file << tem_mov[i][j] << ' ' ;
		}
	      output_file << endl;
	    }
	  for(j=0; j < Nz; j++)
	    {
	      for(i=0; i < Nx; i++)
		{
		  output_file << psi_mov[i][j] << ' ' ;
		}
	      output_file << endl;
	    }
	  // close the file
	  output_file.close();
	}

      // Begin loops to compute derivatives
      for(j=1; j < Nz-1; j++)
	{
	  for(i=0; i < NN; i++)
	    {
	      // Compute derivatives HERE
	      if(i>0)
		{
		  tmp_var = 0.;
		  for(l=1; l < i; l++)
		    {
		      m = i - l;
		      tmp_var += (-(psi_v[m][j+1]-psi_v[m][j-1])/(2.*dz))*ome_v[l][j]*((double) l)
			+psi_v[m][j]*((double) m)*((ome_v[l][j+1]-ome_v[l][j-1])/(2.*dz));
		    }
		  tmp_var *= (-M_PI / (2.*a));
		  dome_v[i][j] = tmp_var;
		  
		  tmp_var = 0.;
		  for(l=i+1; l < NN; l++)
		    {
		      m = l - i;
		      tmp_var += ((psi_v[m][j+1]-psi_v[m][j-1])/(2.*dz))*ome_v[l][j]*((double) l)
			+ psi_v[m][j]*((double) m)*((ome_v[l][j+1]-ome_v[l][j-1])/(2.*dz))
			- ((psi_v[l][j+1]-psi_v[l][j-1])/(2.*dz))*ome_v[m][j]*((double) m)
			- psi_v[l][j]*((double) l)*((ome_v[m][j+1]-ome_v[m][j-1])/(2.*dz));
		    }
		  tmp_var *= (-M_PI / (2.*a));
		  dome_v[i][j] += tmp_var;
		  dome_v[i][j] += Ra*Pr*((double) i)*M_PI*Tem_v[i][j]/a
		    + Pr*(-pow(((double) i)*M_PI/a,2)*ome_v[i][j]
			  +(ome_v[i][j+1] - 2.*ome_v[i][j] + ome_v[i][j-1])/(dz*dz));
		}

	      dTem_v[i][j] = 0.;
	      tmp_var = 0.;

	      for(l=1; l < i; l++)
		{
		  m = i - l;
		  tmp_var += -((psi_v[m][j+1]-psi_v[m][j-1])/(2.*dz))*Tem_v[l][j]*((double) l)
		    + psi_v[m][j] * ((Tem_v[l][j+1]-Tem_v[l][j-1])/(2.*dz)) * ((double) m);
		}
	      tmp_var *= (-M_PI / (2*a));
	      dTem_v[i][j] = tmp_var;
	      
	      tmp_var = 0.;		
	      for(l=i+1; l<NN;l++)
		{
		  m = l - i;
		  tmp_var += ((psi_v[m][j+1]-psi_v[m][j-1])/(2.*dz))*Tem_v[l][j]*((double) l)
		    + psi_v[m][j] * ((Tem_v[l][j+1] - Tem_v[l][j-1])/(2.*dz))* ((double) m)
		    + ((psi_v[l][j+1] - psi_v[l][j-1])/(2.*dz))*Tem_v[m][j]*((double) m)
		    + psi_v[l][j] * ((Tem_v[m][j+1] - Tem_v[m][j-1])/(2.*dz))*((double) l);
		}
	      tmp_var *= (-M_PI / (2.*a));
	      dTem_v[i][j] += tmp_var;

	      dTem_v[i][j] += -(M_PI/a)*psi_v[i][j]*((Tem_v[0][j+1]-Tem_v[0][j-1])/(2.*dz)) * ((double) i);
	      
	      dTem_v[i][j] += ((double) i) * M_PI * psi_v[i][j] / a
		-pow(((double) i) * M_PI / a,2)*Tem_v[i][j] 
		+((Tem_v[i][j+1]-2.*Tem_v[i][j]+Tem_v[i][j-1])/(dz*dz));
	    }
	}
      // End loops that computed derivatives

      // Increment values & move derivatives along 
      for(j=1; j < Nz - 1; j++)
	{
	  for(i=0; i < NN; i++)
	    {
	      // increment values
	      Tem_v[i][j] += dt * (1.5*dTem_v[i][j]+0.5*dTemO_v[i][j])/2.;
	      ome_v[i][j] += dt * (1.5*dome_v[i][j]+0.5*domeO_v[i][j])/2.;	      
	      // Update old derivatives
	      dTemO_v[i][j] = dTem_v[i][j];
	      domeO_v[i][j] = dome_v[i][j];
	    }
	}
      
      time += dt;
    };
  
  
  
  return 0;
}
